//-------------------------------------------------------------------------------------------------
// ParseBuffer.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "ParseBuffer.hpp"
#include "engineAssert.hpp"


//-------------------------------------------------------------------------------------------------
// Class constants
//-------------------------------------------------------------------------------------------------


//-------------------------------------------------------------------------------------------------
// Constructors
//-------------------------------------------------------------------------------------------------
ParseBuffer::ParseBuffer( const char* buffer, int bufferSize, const char* bufferName )
:
_bufferName( bufferName ),
_buffer( buffer ),
_bufferSize( bufferSize ),
_read( _buffer ),
_lineNum( 1 ),
_lineStart( _buffer )
{
	_SetBufferSize();
}


ParseBuffer::ParseBuffer( const std::string& buffer, int bufferSize, const char* bufferName )
:
_bufferName( bufferName ),
_buffer( buffer.c_str() ),
_bufferSize( bufferSize ),
_read( _buffer ),
_lineNum( 1 ),
_lineStart( _buffer )
{
	_SetBufferSize();
}


//-------------------------------------------------------------------------------------------------
// Destructor
//-------------------------------------------------------------------------------------------------
ParseBuffer::~ParseBuffer()
{
}



//-------------------------------------------------------------------------------------------------
// SetBufferSize
//-------------------------------------------------------------------------------------------------
void ParseBuffer::_SetBufferSize()
{
	if( _bufferSize == -1 )
	{
		_bufferSize = strlen( _buffer );
	}
}



//-------------------------------------------------------------------------------------------------
// StripComments
//-------------------------------------------------------------------------------------------------
void ParseBuffer::StripComments()
{
	VALIDATE_MSG0( _read == _buffer, "ParseBuffer::StripComments() : Cannot strip comments after parsing has begun!" );

	/// Allocate a duplicate of the buffer
	char* strippedBuffer = new char[ _bufferSize + 1 ];
	VALIDATE( strippedBuffer );
	strippedBuffer[ 0 ] = '\0';

	/// Set up read / write pointers
	const char* read = _buffer;
	char* write = strippedBuffer;

	/// Make use of a miniature state machine
	enum CommentParseState
	{
		COMMENT_PARSE_STATE_NORMAL,
		COMMENT_PARSE_STATE_COMMENT_LINE,
		COMMENT_PARSE_STATE_COMMENT_BLOCK,
	};

	CommentParseState parseState = COMMENT_PARSE_STATE_NORMAL;

	/// Copy the buffer into <strippedBuffer>, omitting commented-out portions
	while( *read )
	{
		if( parseState == COMMENT_PARSE_STATE_NORMAL )
		{
			/// Search for next '/' (start of line or block comment)
			const char* scan = strchr( read, '/' );
			if( !scan )
			{
				/// Copy the remaining buffer and quit!
				*write = '\0';
				strcat( strippedBuffer, read );
				write += strlen( read );
				break;
			}

			/// Copy everything up to the '/'
			_CopyUpToScan( read, write, scan );

			/// Determine which comment type (if any) the slash introduces
			char nextChar = read[ 1 ];
			if( nextChar == '/' )
			{
				/// Enter line comment
				parseState = COMMENT_PARSE_STATE_COMMENT_LINE;
				read += 2;
				continue;
			}
			else if( nextChar == '*' )
			{
				/// Enter block comment
				parseState = COMMENT_PARSE_STATE_COMMENT_BLOCK;
				read += 2;
				continue;
			}
			else
			{
				/// False alarm; this is just an ordinary slash; copy it and continue
				*write = *read;
				write ++;
				read ++;
			}
		}
		else if( parseState == COMMENT_PARSE_STATE_COMMENT_LINE )
		{
			/// Skip ahead to EOL (or EOF)
			size_t commentedOutCharacters = strcspn( read, "\r\n" );
			read += commentedOutCharacters;
			if( *read == '\r' )
			{
				*write = *read;
				write ++;
				read ++;
			}

			if( *read == '\n' )
			{
				*write = *read;
				write ++;
				read ++;
			}

			parseState = COMMENT_PARSE_STATE_NORMAL;
		}
		else if( parseState == COMMENT_PARSE_STATE_COMMENT_BLOCK )
		{
			for( ;; )
			{
				char nextChar = read[ 1 ];
				if( *read == '\0' )
				{
					break;
				}
				else if( *read == '*' && nextChar == '/' )
				{
					/// End of comment block
					parseState = COMMENT_PARSE_STATE_NORMAL;
					read += 2;
					break;
				}
				else if( *read == '\r' || *read == '\n' )
				{
					/// Preserve newline characters (only)
					*write = *read;
					write ++;
					read ++;
				}
				else
				{
					read ++;
				}
			}

			*write = '\0';
		}
	}

	/// Copy the stripped buffer onto the old buffer
	strcpy( const_cast<char*>( _buffer ), strippedBuffer );
	_bufferSize = strlen( _buffer );

	/// Free the old buffer
	delete[] strippedBuffer;
}



//-------------------------------------------------------------------------------------------------
// CopyUpToScan
//-------------------------------------------------------------------------------------------------
void ParseBuffer::_CopyUpToScan( const char*& read, char*& write, const char* scan )
{
	/// Compute byte-size of copy
	int copyLength = scan - read;
	if( !copyLength )
		return;

	strncpy( write, read, copyLength );
	write += copyLength;
	read += copyLength;
	*write = '\0';
}



//-------------------------------------------------------------------------------------------------
// ParseInt
//-------------------------------------------------------------------------------------------------
int ParseBuffer::ParseInt()
{
	SkipWhitespace();

	const int MAX_STRING_LEN = 1024;
	char parseString[ MAX_STRING_LEN ];
	char* write = parseString;
	while( isdigit( *_read ) || *_read == '-' || *_read == '+' )
	{
		*write = *_read;
		write ++;
		_read ++;

		if( write - parseString >= MAX_STRING_LEN )
		{
			break;
		}
	}
	*write = '\0';

	/// Check for syntax error
	if( write == parseString )
	{
		_SyntaxError( "couldn't find expected integer expression" );
	}

	return atoi( parseString );
}



//-------------------------------------------------------------------------------------------------
// ParseFloat
//-------------------------------------------------------------------------------------------------
float ParseBuffer::ParseFloat()
{
	return (float) ParseDouble();
}



//-------------------------------------------------------------------------------------------------
// ParseDouble
//-------------------------------------------------------------------------------------------------
double ParseBuffer::ParseDouble()
{
	SkipWhitespace();

	const int MAX_STRING_LEN = 1024;
	char parseString[ MAX_STRING_LEN ];
	char* write = parseString;
	while( isdigit( *_read ) || *_read == '.' || *_read == '-' || *_read == '+' )
	{
		*write = *_read;
		write ++;
		_read ++;

		if( write - parseString >= MAX_STRING_LEN - 1 )
		{
			break;
		}
	}
	*write = '\0';

	/// Check for syntax error
	if( write == parseString )
	{
		_SyntaxError( "couldn't find expected floating point expression" );
	}

	return atof( parseString );
}


//-------------------------------------------------------------------------------------------------
// ParseQuotedString
//-------------------------------------------------------------------------------------------------
void ParseBuffer::ParseQuotedString( char* destBuffer, int destBufferSize, bool allowLineWrap )
{
	char* write = destBuffer;

	ParseExpectedToken( '\"' );
	while( *_read && *_read != '\"' )
	{
		if( !allowLineWrap && (*_read == '\r' || *_read == '\n') )
		{
			_SyntaxError( "illegal EOL found inside quoted string" );
		}

		if( write - destBuffer >= destBufferSize - 1 )
		{
			break;
		}

		*write = *_read;
		write ++;
		_read ++;
	}

	*write = '\0';

	if( *_read == '\"' )
	{
		/// Skip closing quotation
		_read ++;
	}
	else if( !*_read )
	{
		_SyntaxError( "illegal EOF found inside quoted string" );
	}
	else
	{
		char syntaxErrorMessage[ 1024 ];
		sprintf( syntaxErrorMessage, "quoted string was too long, overflowed buffer size (max len = %d characters)", destBufferSize - 1 );
		_SyntaxError( syntaxErrorMessage );
	}
}


//-------------------------------------------------------------------------------------------------
// ParseQuotedString
//-------------------------------------------------------------------------------------------------
void ParseBuffer::ParseQuotedString( std::string& destBuffer, int destBufferMaxSize, bool allowLineWrap )
{
	PANIC( "ParseBuffer::ParseQuotedString( std::string& destBuffer, ... ) : Sorry, not yet implemented." );
}


//-------------------------------------------------------------------------------------------------
// ParseContiguousString
//-------------------------------------------------------------------------------------------------
void ParseBuffer::ParseContiguousString( char* destBuffer, int destBufferSize )
{
	char* write = destBuffer;

	SkipWhitespace();
	while( *_read && !isspace( *_read ) )
	{
		if( write - destBuffer >= destBufferSize - 1 )
		{
			break;
		}

		*write = *_read;
		write ++;
		_read ++;
	}

	*write = '\0';

	if( write - destBuffer >= destBufferSize - 1 )
	{
		char syntaxErrorMessage[ 1024 ];
		sprintf( syntaxErrorMessage, "continguous string was too long, overflowed buffer size (max len = %d characters)", destBufferSize - 1 );
		_SyntaxError( syntaxErrorMessage );
	}
}


//-------------------------------------------------------------------------------------------------
// ParseStringToEndOfLine
//-------------------------------------------------------------------------------------------------
void ParseBuffer::ParseStringToEndOfLine( char* destBuffer, int destBufferSize, bool removeTrailingWhitespace )
{
	char* write = destBuffer;

	SkipWhitespace();

	while( *_read && *_read != '\r' && *_read != '\n' )
	{
		if( write - destBuffer >= destBufferSize - 1 )
		{
			break;
		}

		*write = *_read;
		write ++;
		_read ++;
	}

	*write = '\0';

	if( write - destBuffer >= destBufferSize - 1 )
	{
		char syntaxErrorMessage[ 1024 ];
		sprintf( syntaxErrorMessage, "string-until-end-of-line was too long, overflowed buffer size (max len = %d characters)", destBufferSize - 1 );
		_SyntaxError( syntaxErrorMessage );
	}

	if( removeTrailingWhitespace )
	{
		int len = strlen( destBuffer );
		while( len > 0 && isspace( destBuffer[ len - 1 ] ) )
		{
			destBuffer[ len - 1 ] = '\0';
			len --;
		}
	}
}

//-------------------------------------------------------------------------------------------------
// ParseVector2
//-------------------------------------------------------------------------------------------------
Vector2 ParseBuffer::ParseVector2()
{
	float x, y;
	bool usingParens = false;

	char token = AdvanceToNextToken();
	if( token == '(' )
	{
		ParseExpectedToken( '(' );
		usingParens = true;
	}

	x = ParseFloat();
	ParseExpectedToken( ',' );
	y = ParseFloat();

	if( usingParens )
	{
		ParseExpectedToken( ')' );
	}

	return Vector2( x, y );
}


//-------------------------------------------------------------------------------------------------
// ParseVector3
//-------------------------------------------------------------------------------------------------
Vector3 ParseBuffer::ParseVector3()
{
	float x, y, z;
	bool usingParens = false;

	char token = AdvanceToNextToken();
	if( token == '(' )
	{
		ParseExpectedToken( '(' );
		usingParens = true;
	}

	x = ParseFloat();
	ParseExpectedToken( ',' );
	y = ParseFloat();
	ParseExpectedToken( ',' );
	z = ParseFloat();

	if( usingParens )
	{
		ParseExpectedToken( ')' );
	}

	return Vector3( x, y, z );
}


//-------------------------------------------------------------------------------------------------
// ParseVector4
//-------------------------------------------------------------------------------------------------
Vector4 ParseBuffer::ParseVector4()
{
	float x, y, z, t;
	bool usingParens = false;

	char token = AdvanceToNextToken();
	if( token == '(' )
	{
		ParseExpectedToken( '(' );
		usingParens = true;
	}

	x = ParseFloat();
	ParseExpectedToken( ',' );
	y = ParseFloat();
	ParseExpectedToken( ',' );
	z = ParseFloat();
	ParseExpectedToken( ',' );
	t = ParseFloat();

	if( usingParens )
	{
		ParseExpectedToken( ')' );
	}

	return Vector4( x, y, z, t );
}


//-------------------------------------------------------------------------------------------------
// ParseExpectedToken
//-------------------------------------------------------------------------------------------------
void ParseBuffer::ParseExpectedToken( char token )
{
	/// Skip to first token
	SkipWhitespace();

	/// Make sure token matches expectations
	if( *_read != token )
	{
		char syntaxErrorMessage[ 1024 ];
		sprintf( syntaxErrorMessage, "couldn't find expected ' %c ' character", token );
		_SyntaxError( syntaxErrorMessage );
	}

	/// Skip past token
	_read ++;
}


//-------------------------------------------------------------------------------------------------
// ParseExpectedToken
//-------------------------------------------------------------------------------------------------
void ParseBuffer::ParseExpectedToken( const char* token )
{
	int tokenLen = strlen( token );

	/// Skip to first token
	SkipWhitespace();

	/// Make sure token matches expectations
	if( strnicmp( token, _read, tokenLen ) )
	{
		char syntaxErrorMessage[ 1024 ];
		sprintf( syntaxErrorMessage, "couldn't find expected token-word \"%s\"", token );
		_SyntaxError( syntaxErrorMessage );
	}

	/// Skip past token
	_read += tokenLen;
}


//-------------------------------------------------------------------------------------------------
// AdvanceToNextToken
//-------------------------------------------------------------------------------------------------
char ParseBuffer::AdvanceToNextToken()
{
	SkipWhitespace();
	return *_read;
}


//-------------------------------------------------------------------------------------------------
// SyntaxError
//-------------------------------------------------------------------------------------------------
void ParseBuffer::_SyntaxError( const char* errorDesc )
{
	/// Get the text for the line on which the error occurred
	const int MAX_STRING_LEN = 1024;
	char line[ MAX_STRING_LEN ];
	strncpy( line, _lineStart, MAX_STRING_LEN - 1 );
	line[ MAX_STRING_LEN - 1 ] = '\0';
	char* scan = strpbrk( line, "\r\n" );
	if( scan )
	{
		*scan = '\0';
	}

	/// Get the text which occurred instead of (whatever was expected)
	char syntaxError[ MAX_STRING_LEN ];
	strncpy( syntaxError, _read, MAX_STRING_LEN - 1 );
	syntaxError[ MAX_STRING_LEN - 1 ] = '\0';
	scan = strpbrk( syntaxError, "\r\n" );
	if( scan )
	{
		*scan = '\0';
	}

	PANIC_MSG5( "Syntax error on line %d of buffer \"%s\":\n\nERROR: %s\n\nOCCURRED AT: \"%s\"\n\nLINE WAS: \"%s\"",
		_lineNum, _bufferName, errorDesc, syntaxError, line );
}



