//-------------------------------------------------------------------------------------------------
// TheApplication.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "TheApplication.hpp"
#include "TheConsole.hpp"
#include "TheMessageHandler.hpp"
#include "TheVideoSystem.hpp"
#include "TheRenderer.hpp"
#include "TheNetwork.hpp"
#include "TheGameInterfaceAbstract.hpp"
#include "TheSoundSystem.hpp"
#include "Timer.hpp"
#include "engineAssert.hpp"
using namespace std;


//-------------------------------------------------------------------------------------------------
// Subsystem name macro (for assertions, etc)
//-------------------------------------------------------------------------------------------------
#define SUBSYSTEM_NAME "TheApplication"


//-------------------------------------------------------------------------------------------------
// Global subsystem instances
//-------------------------------------------------------------------------------------------------
TheApplication theApplication;
TheConsole theConsole;
TheInputSystem theInputSystem;
TheMessageHandler theMessageHandler;
TheVideoSystem theVideoSystem;
TheSoundSystem theSoundSystem;
TheRenderer theRenderer;
TheNetwork theNetwork;


//-------------------------------------------------------------------------------------------------
// Static member initialization
//-------------------------------------------------------------------------------------------------
STATIC bool				TheApplication::_isInitialized		= false;
STATIC string			TheApplication::_appName( "UNNAMED APPLICATION : Call theApplication.SetName() from PreInit() to set name!" );
STATIC string			TheApplication::_baseDir( "./" );


//-------------------------------------------------------------------------------------------------
// constructor
//-------------------------------------------------------------------------------------------------
TheApplication::TheApplication()
:
_cmdLine( "" ),
_hInstance( NULL ),
_isExiting( false ),
_isShutdownRequested( false ),
_frameNumber( 0 ),
_secondsSinceLastFrame( 0.1f ),
_timeFrameBegan( Timer::GetSysTimeSeconds() ),
_fpsAsymptotic( 60.0f ),
_hasFocus( true )
{
}


//-------------------------------------------------------------------------------------------------
// destructor
//-------------------------------------------------------------------------------------------------
TheApplication::~TheApplication()
{
	/// Make sure the system was shut down before being destroyed
	VALIDATE_MSG0( !_isInitialized, SUBSYSTEM_NAME" was destroyed while running; RequestShutdown() should be called first!" );
}


//-------------------------------------------------------------------------------------------------
// Startup
//-------------------------------------------------------------------------------------------------
void TheApplication::Startup()
{
	theConsole.Printf( "::: Initialization : Application (starting)\n" );

	/// Make sure the system doesn't get initialized twice
	VALIDATE_MSG0( !_isInitialized, SUBSYSTEM_NAME" was initialized twice; either a second "SUBSYSTEM_NAME" was started or Startup() was called on an already-running system." );

	/// Make sure an instance of the game already exists
	VALIDATE_MSG0( TheGameInterfaceAbstract::IsGamePresent(),
		"ENGINE USAGE ERROR:\n"
		"  You must create a static instance of a class derived from TheGameInterfaceAbstract.\n"
		"  This derived class - typically named TheGameInterface - serves as the main interface"
		"    by which the engine can call into game code.\n"
		"  See TheGameInterfaceAbstract.hpp and engine documentation for more details.\n" );

	/// Process command line parameters
	_ProcessCommandLine();

	/// Set application initialization complete
	_isInitialized = true;

	/// Inform the game that subsystem initialization is about to begin
	TheGameInterfaceAbstract::Singleton()->PreInit();

	/// Initialize subsystems
	theConsole.Startup();
	theInputSystem.Startup();
	theMessageHandler.Startup();
	theVideoSystem.Startup();
	theSoundSystem.Startup();
	theRenderer.Startup();
//	theNetwork.Startup();

	/// Inform the game that subsystem initialization has finished
	TheGameInterfaceAbstract::Singleton()->PostInit();

	theConsole.Printf( "::: Initialization : Application (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// _Shutdown
//-------------------------------------------------------------------------------------------------
void TheApplication::_Shutdown()
{
	theConsole.Printf( "::: _Shutdown : Application (starting)\n" );

	/// Enter "shutting down" mode
	_isExiting = true;

	/// Shut down the game
	TheGameInterfaceAbstract::Singleton()->Shutdown();

	/// Shut down subsystems
//	theNetwork.Shutdown();
	theRenderer.Shutdown();
	theSoundSystem.Shutdown();
	theVideoSystem.Shutdown();
	theMessageHandler.Shutdown();
	theInputSystem.Shutdown();
	theConsole.Shutdown();

	/// De-initialize the system
	_isInitialized = false;
	theConsole.Printf( "::: _Shutdown : Application (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// CheckForShutdownRequest
//-------------------------------------------------------------------------------------------------
void TheApplication::CheckForShutdownRequest()
{
	if( _isShutdownRequested )
	{
		_Shutdown();
	}
}


//-------------------------------------------------------------------------------------------------
// RequestShutdown
//-------------------------------------------------------------------------------------------------
void TheApplication::RequestShutdown()
{
	_isShutdownRequested = true;
}


//-------------------------------------------------------------------------------------------------
// OnQuitMessage
//-------------------------------------------------------------------------------------------------
void TheApplication::OnQuitMessage()
{
//	ASSERT( _isInitialized );
	RequestShutdown();
}


//-------------------------------------------------------------------------------------------------
// RunFrame
//-------------------------------------------------------------------------------------------------
void TheApplication::RunFrame()
{
	ASSERT_MSG0( _isInitialized, SUBSYSTEM_NAME"::RunFrame() was called while "SUBSYSTEM_NAME" was uninitialized; call Startup() first." );

	/// Don't do any work if a shutdown is in progress
	if( IsExiting() )
		return;

	/// Update the frame counter, etc.
	_UpdateFrameInfo();

	/// Update the sound engine for this frame
	theSoundSystem.Update();

	/// Update the input for this frame
	theInputSystem.Update();

	/// Tell the game to run (if it exists)
	TheGameInterfaceAbstract::Singleton()->RunFrame();

	/// Tell the renderer to draw everything
	theRenderer.RenderScene();

	/// Swap front / back drawing buffers
	theVideoSystem.SwapBuffers();

	/// Reset the screen
	theRenderer.ClearScreen( Rgba::BLACK );
	theRenderer.ResetView();

	/// Check if there have been any shutdown requests
	CheckForShutdownRequest();
}


//-------------------------------------------------------------------------------------------------
// _ProcessCommandLine
//-------------------------------------------------------------------------------------------------
void TheApplication::_ProcessCommandLine()
{
	/// Allow the game first pass at the command line; let it change the string if it wants
	TheGameInterfaceAbstract::Singleton()->ProcessCommandLine( _cmdLine );
}



//-------------------------------------------------------------------------------------------------
// _UpdateFrameInfo
//-------------------------------------------------------------------------------------------------
void TheApplication::_UpdateFrameInfo()
{
	/// Advance the frame counter
	_frameNumber ++;

	/// Calculate elapsed time for the previous frame
	double timeNow = Timer::GetSysTimeSeconds();
	_secondsSinceLastFrame = timeNow - _timeFrameBegan;
	_timeFrameBegan = timeNow;

	/// Clamp the time to a reasonable, nonzero range
	_secondsSinceLastFrame = ClampDouble( _secondsSinceLastFrame, 0.001, 1.0 );

	/// Calculate instantaneous framerate
	float fpsInstantaneous = 1.0f / (float) _secondsSinceLastFrame;
	_fpsAsymptotic *= 0.90f;
	_fpsAsymptotic += (0.10f * fpsInstantaneous);
}


//-------------------------------------------------------------------------------------------------
// OnActivate
//-------------------------------------------------------------------------------------------------
void TheApplication::OnAppGainFocus()
{
	_hasFocus = true;

	/// Tell the input system to override mouse cursor hide + lock
	theInputSystem.OnAppGainFocus();
}


//-------------------------------------------------------------------------------------------------
// OnDeactivate
//-------------------------------------------------------------------------------------------------
void TheApplication::OnAppLoseFocus()
{
	_hasFocus = false;

	/// Tell the input system to remove mouse cursor hide/lock override
	theInputSystem.OnAppLoseFocus();
}


