//-------------------------------------------------------------------------------------------------
// Entity_EnemyGrunt.cpp
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "engineAssert.hpp"
#include "Entity.hpp"
#include "Particle.hpp"
#include "TheGame.hpp"


const double GRUNT_TURN_DEGREES_PER_SECOND = 30.0;
const double GRUNT_MOVE_TILES_PER_SECOND = 2.0;
const double GRUNT_MOVE_TILES_PER_SECOND_WHILE_STUNNED = 1.5;
const double GRUNT_CHARGE_TILES_PER_SECOND = 5.0;
const double GRUNT_LIFESPAN_SECONDS = 30.0;
const float GRUNT_CHARGE_PLAYER_PROXIMITY = 3.0f;
const float GRUNT_CHARGE_HURT_PROXIMITY = 0.5f;
const double GRUNT_CHARGE_WAIT_SECONDS = 0.0;
const double GRUNT_PAUSE_AFTER_CHARGE_SECONDS = 0.0;
const float GRUNT_CHARGE_DAMAGE_PER_SECOND = 25.0f;
const float GRUNT_STUN_TIME = 4.0f;


enum // grunt states
{
	GRUNT_STATE_WANDER,
	GRUNT_STATE_CHARGE_WAIT,
	GRUNT_STATE_CHARGE_GO,
	GRUNT_STATE_PAUSE_AFTER_CHARGE,
	GRUNT_STATE_STUNNED,
};


//-------------------------------------------------------------------------------------------------
// Spawn_EnemyGrunt
//-------------------------------------------------------------------------------------------------
void Entity::Spawn_EnemyGrunt()
{
	m_anim = new Animation( "data/gfx/Enemies/Knight.tga" );
	m_bounds.SetFromMinXYMaxXY( -8.0f, -8.0f, 8.0f, 8.0f );
	m_rotVel = RandomFloat( (float)GRUNT_TURN_DEGREES_PER_SECOND, -(float)GRUNT_TURN_DEGREES_PER_SECOND );
	m_genericStateEnum = GRUNT_STATE_WANDER;
	m_hitPoints = 10.0f * ( ( theGame->m_player->m_difficulty * theGame->m_player->m_difficulty * 0.1 ) * 0.5 );
	if( m_hitPoints < 10 )
	{
		m_hitPoints = 10;
	}
	m_damageGiveMultiplier = theGame->m_player->m_difficulty * 0.25;
	if( m_damageGiveMultiplier < 1 )
	{
		m_damageGiveMultiplier = 1;
	}
	m_damageTakeMultiplier = theGame->m_player->m_difficulty * 0.25;
	if( m_damageTakeMultiplier < 1 )
	{
		m_damageTakeMultiplier = 1;
	}
	m_drawRotated = false;
	m_isHitByPlayerBullets = true;
	m_isEnemy = true;
	double timeNow = theApplication.GetTimeFrameBegan();
}


//-------------------------------------------------------------------------------------------------
// Think_EnemyGrunt
//-------------------------------------------------------------------------------------------------
void Entity::Think_EnemyGrunt()
{
	double timeNow = theApplication.GetTimeFrameBegan();
	
	if( m_stunned == true && m_genericStateEnum != GRUNT_STATE_STUNNED )
	{
		m_genericStateEnum = GRUNT_STATE_STUNNED;
		m_genericTimeStamp = timeNow + GRUNT_STUN_TIME;
	}

	if( m_genericStateEnum == GRUNT_STATE_WANDER )
	{
		//-------------------------------------------------------------------------------------------------
		// WANDER
		//-------------------------------------------------------------------------------------------------
		m_color = Rgba( 255, 255, 255, 255 );

		/// Charge the player if he is nearby
		if( !theGame->m_player->IsInvisible() && theGame->IsPointWithinGivenRadiusOfPlayer( m_pos, GRUNT_CHARGE_PLAYER_PROXIMITY ) )
		{
			/// Charge the player if he is nearby
			m_genericStateEnum = GRUNT_STATE_CHARGE_WAIT;
			m_genericTimeStamp = timeNow + GRUNT_CHARGE_WAIT_SECONDS;
		}
		else
		{
			/// Wander
			const double turnSpeed = m_rotVel * theApplication.GetFrameSeconds();
			const double moveSpeed = GRUNT_MOVE_TILES_PER_SECOND * theApplication.GetFrameSeconds();

			/// Constant rotation
			m_rot += (float) turnSpeed;
			m_rot = fmod( m_rot, 360.0 );

			/// Build move vector
			Vector2 moveVector;
			moveVector.SetFromRadiusCosSinDegrees( moveSpeed, m_rot );
			bool success = TryToMove( moveVector );
			if( !success )
			{
				m_rot += 30.0f;
			}
		}
	}
	else if( m_genericStateEnum == GRUNT_STATE_CHARGE_WAIT )
	{
		//-------------------------------------------------------------------------------------------------
		// WAITING TO CHARGE
		//-------------------------------------------------------------------------------------------------

		if( timeNow > m_genericTimeStamp )
		{
			Vector2 vectorToPlayer = theGame->m_player->pos - m_pos;
			if( theGame->m_player->IsInvisible() )
			{
				m_rot = RandomFloat( 0.0f, 360.0f );
			}
			else
			{
				m_rot = vectorToPlayer.GetDegrees();
			}
			m_genericStateEnum = GRUNT_STATE_CHARGE_GO;
		}
	}
	else if( m_genericStateEnum == GRUNT_STATE_CHARGE_GO )
	{
		//-------------------------------------------------------------------------------------------------
		// CHARGING!
		//-------------------------------------------------------------------------------------------------

		const double moveSpeed = GRUNT_CHARGE_TILES_PER_SECOND * theApplication.GetFrameSeconds();
		Vector2 moveVector;
		moveVector.SetFromRadiusCosSinDegrees( moveSpeed, m_rot );
		bool success = TryToMove( moveVector );
		if( success )
		{
			/// See if we're trampling the player
			if( theGame->IsPointWithinGivenRadiusOfPlayer( m_pos, GRUNT_CHARGE_HURT_PROXIMITY ) )
			{
				theGame->m_player->AdjustHealth( ( ( GRUNT_CHARGE_DAMAGE_PER_SECOND * m_damageGiveMultiplier ) * (float)theApplication.GetFrameSeconds() ) * -1.0f );
			}
		}
		else
		{
			/// We hit something; stop charging now
			m_genericStateEnum = GRUNT_STATE_PAUSE_AFTER_CHARGE;
			m_genericTimeStamp = timeNow + GRUNT_PAUSE_AFTER_CHARGE_SECONDS;
		}
	}
	else if( m_genericStateEnum == GRUNT_STATE_PAUSE_AFTER_CHARGE )
	{
		if( timeNow > m_genericTimeStamp )
		{
			/// Done being paused; return to wander mode
			m_genericStateEnum = GRUNT_STATE_WANDER;
		}
	}
	else if( m_genericStateEnum == GRUNT_STATE_STUNNED )
	{
		//-------------------------------------------------------------------------------------------------
		// STUNNED
		//-------------------------------------------------------------------------------------------------

		/// Wander around stunned
		const double turnSpeed = m_rotVel * theApplication.GetFrameSeconds();
		const double moveSpeed = GRUNT_MOVE_TILES_PER_SECOND_WHILE_STUNNED * theApplication.GetFrameSeconds();

		m_color = Rgba( 255, 255, 128, 255 );

		/// Constant rotation
		m_rot += (float) turnSpeed;
		m_rot = fmod( m_rot, 360.0 );

		/// Build move vector
		Vector2 moveVector;
		moveVector.SetFromRadiusCosSinDegrees( moveSpeed, m_rot );
		bool success = TryToMove( moveVector );
		if( !success )
		{
			m_rot += 30.0f;
		}

		/// See if it's time to become unstunned
		if( timeNow > m_genericTimeStamp )
		{
			m_genericStateEnum = GRUNT_STATE_WANDER;
			m_stunned = false;
		}
	}
	else
	{
		m_genericStateEnum = GRUNT_STATE_WANDER;
	}

	/// See if it's time to die
	double age = timeNow - m_birthTime;
	if( age > GRUNT_LIFESPAN_SECONDS )
	{
		m_readyToDie = true;
	}
}


//-------------------------------------------------------------------------------------------------
// Die_EnemyGrunt
//-------------------------------------------------------------------------------------------------
void Entity::Die_EnemyGrunt()
{
	int randomItemSpawn = RandomInt( 0 , 11 );
	if( randomItemSpawn == 1 )
	{
		SpawnNewEntity( ENT_ITEMPICKUP_BOMB, m_pos );
	}
	else if( randomItemSpawn == 2 )
	{
		SpawnNewEntity( ENT_ITEMPICKUP_HEALTH, m_pos );
	}

	delete m_anim ;

	Particle* p = Particle::Spawn();
	p->m_pos = m_pos;
	p->m_startAlpha = 1.0f;
	p->m_endAlpha = 0.0f;
	p->m_startScale = 1.0f;
	p->m_endScale = 5.0f;
	p->m_timeToLive = 0.3;
	p->m_anim = new Animation( "data/gfx/particles/KnightExplode.tga" );

}



