//-------------------------------------------------------------------------------------------------
// MapData.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "MapData.hpp"
#include "engineAssert.hpp"
using namespace std;


//-------------------------------------------------------------------------------------------------
// Static member initialization
//-------------------------------------------------------------------------------------------------
STATIC std::vector<MapData*>					MapData::s_mapDataList;


//-------------------------------------------------------------------------------------------------
// Constructors
//-------------------------------------------------------------------------------------------------
MapData::MapData( const Image& mapDataImage, const char* mapName )
{
	/// Validate size
	VALIDATE_MSG4( mapDataImage.GetHeight() == MAP_TILE_SIZE_X, "Map data file \"%s\" was not %dx%d; illegal height %d", mapName, MAP_TILE_SIZE_X, MAP_TILE_SIZE_Y, mapDataImage.GetHeight() );
	VALIDATE_MSG4( mapDataImage.GetWidth() == MAP_TILE_SIZE_Y, "Map data file \"%s\" was not %dx%d; illegal width %d", mapName, MAP_TILE_SIZE_X, MAP_TILE_SIZE_Y, mapDataImage.GetWidth() );

	/// Scan through map data pixels and parse each one (first pass)
	int x, y;
	for( x = 0; x < MAP_TILE_SIZE_X; x ++ )
	for( y = 0; y < MAP_TILE_SIZE_Y; y ++ )
	{
		Rgba pixelColor = mapDataImage.GetPixelColor( x, y );
		
		/// Find a material in the materials array that matches this color
		Material& mat = Material::FindMaterialByColor( pixelColor );
		if( mat.m_id == -1 )
		{
			char errorMessage[ 1024 ];
			sprintf( errorMessage, "Error parsing level data for map \"%s\" - unrecognized material color (%03d %03d %03d) at coordinates (%d, %d)!\n",
				mapName, pixelColor.r, pixelColor.g, pixelColor.b, x, y );

			ASSERT_MSG1( mat.m_id != -1, "%s\n", errorMessage );
		}

		/// Assign that material's ID
		tileArray[ x ][ y ].m_materialID = mat.m_id;
		tileArray[ x ][ y ].m_debugFrameCountMark = 0;
	}
}


//-------------------------------------------------------------------------------------------------
// Destructor
//-------------------------------------------------------------------------------------------------
MapData::~MapData()
{
}


//-------------------------------------------------------------------------------------------------
// LoadAllMaps
//-------------------------------------------------------------------------------------------------
STATIC void MapData::LoadAllMaps()
{
	theConsole.Printf( "Loading all maps...\n" );

	/// Load as many maps as exist
	bool finished = false;
	for( int mapNumber = 1; !finished ; mapNumber ++ )
	{
		VALIDATE( 0 );
		//finished = !LoadMap( mapNumber );
	}

	VALIDATE_MSG0( s_mapDataList.size(), "MapData::LoadAllMaps() : ERROR : no maps were loaded!\n" );
}


//-------------------------------------------------------------------------------------------------
// LoadMap
//-------------------------------------------------------------------------------------------------
STATIC bool MapData::LoadMap( const char* const mapSet, const int mapNumber )
{
	char mapName[ 1024 ];

	/// Silently attempt to open the file for reading; it's okay if we fail
	sprintf( mapName, "data/maps/%s%03d.tga", mapSet, mapNumber );
	FILE* file = fopen( mapName, "rb" );
	if( !file )
	{
		/// Couldn't load; signal failure return
		return false;
	}

	/// Close the file (only open for testing purposes)
	fclose( file );

	Image mapDataImage( mapName );
	MapData* mapData = new MapData( mapDataImage, mapName );

	// load metadata
	mapData->ParseMetaData( mapSet, mapNumber );

	// spawn entities
	s_mapDataList.push_back( mapData );
	mapData->SpawnAllEntities();

	return true;
}

//-------------------------------------------------------------------------------------------------
void MapData::ParseMetaData( const char* const mapSet, const int mapNumber )
{
	char mapName[ 1024 ];
	sprintf( mapName, "data/maps/%s%03d.txt", mapSet, mapNumber );
	ParseFile parseFile( mapName );
	parseFile.StripComments();
	parseFile.ParseExpectedToken( "lightScale" );
	m_lightScale = parseFile.ParseFloat();
	parseFile.ParseExpectedToken( "fogColor" );
	m_fogColor.r = parseFile.ParseInt();
	m_fogColor.g = parseFile.ParseInt();
	m_fogColor.b = parseFile.ParseInt();
	m_fogColor.a = 255;
}

//-------------------------------------------------------------------------------------------------
// GetMaterialForMapXY
//-------------------------------------------------------------------------------------------------
Material& MapData::GetMaterialForMapXY( int x, int y )
{
	int materialID = tileArray[ x ][ y ].m_materialID;
	if( materialID < 0 )
	{
		ASSERT_MSG2( materialID >= 0, "GetMaterialForMapXY( %d, %d ) failed; unknown material type!\n", x, y );
		return Material::s_missingMaterial;
	}

	return g_materialTable[ materialID ];
}



///-------------------------------------------------------------------------------------------------
/// FindFirstMaterialByID
///-------------------------------------------------------------------------------------------------
IntVector2 MapData::FindFirstTileWithMaterialID( int materialID )
{
	for (int x = 0; x < MAP_TILE_SIZE_X; x++ )
	for ( int y = 0; y < MAP_TILE_SIZE_Y; y++ )
	{
		Material& material = GetMaterialForMapXY( x, y );
		if ( material.m_id	== materialID )
		{
			return IntVector2( x, y );
		}
	}
	return IntVector2( -1, -1 );
}



///------------------------------------------------------------------------------------------------
/// ChangeMaterialAtPosition
///------------------------------------------------------------------------------------------------
void MapData::ChangeMaterialAtPosition( const IntVector2 &tilePos, const std::string& materialName )
{
	theConsole.Printf( "Changing material at %d,%d to \"%s\"\n", tilePos.x, tilePos.y, materialName.c_str() );
	Material& material = Material::FindMaterialByName( materialName.c_str() );
	tileArray[ tilePos.x ][ tilePos.y ].m_materialID = material.m_id ;
}



//-------------------------------------------------------------------------------------------------
// SpawnAllEntities
//-------------------------------------------------------------------------------------------------
void MapData::SpawnAllEntities()
{
	for( int x = 0; x < MAP_TILE_SIZE_X; x++ )
	for( int y = 0; y < MAP_TILE_SIZE_Y; y++ )
	{
		const int materialId = tileArray[ x ][ y ].m_materialID;
		Material& mat = g_materialTable[ materialId ];

		if( mat.m_spawnEntityType )
		{
			if( mat.m_spawnEntityType == ENT_GENERATOR_WIZROBE )
			{
				mat.m_spawnEntityType = ENT_GENERATOR_WIZROBE;
			}
			else if( mat.m_spawnEntityType == ENT_GENERATOR_BOSS1 )
			{
				mat.m_spawnEntityType = ENT_GENERATOR_BOSS1;
			}
			else if( mat.m_spawnEntityType == ENT_GENERATOR_SINGLEBOSS1 )
			{
				mat.m_spawnEntityType = ENT_GENERATOR_SINGLEBOSS1;
			}
			else if( mat.m_spawnEntityType == ENT_GENERATOR_BOSS2 )
			{
				mat.m_spawnEntityType = ENT_GENERATOR_BOSS2;
			}
			else if( mat.m_spawnEntityType == ENT_GENERATOR_SINGLEBOSS2 )
			{
				mat.m_spawnEntityType = ENT_GENERATOR_SINGLEBOSS2;
			}
			else if( mat.m_spawnEntityType == ENT_GENERATOR_MUSHROOM )
			{
				mat.m_spawnEntityType = ENT_GENERATOR_MUSHROOM;
			}

			Entity* e = new Entity( mat.m_spawnEntityType );
			e->m_pos.SetXY( (float)x + 0.5f, (float)y + 0.5f );
			entityList.push_back( e );
			e->SpawnFunc();
			
			// replace material on map with replacement material
			if( mat.m_replacementMaterialOnSpawn )
			{
				tileArray[ x ][ y ].m_materialID = Material::FindMaterialByName( mat.m_replacementMaterialOnSpawn ).m_id;
			}
			
			continue;
		}
	}
}