//-------------------------------------------------------------------------------------------------
// Material.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "Material.hpp"
#include "engineAssert.hpp"
#include "MapData.hpp"
using namespace std;


//-------------------------------------------------------------------------------------------------
// Global data
//-------------------------------------------------------------------------------------------------
Material g_materialTable[ MAX_NUM_MATERIALS ];
int g_numMaterials = 0;
STATIC std::map< std::string, TypeOfEntity >	Material::s_entityTypeMap;


//-------------------------------------------------------------------------------------------------
// Static member initialization
//-------------------------------------------------------------------------------------------------
//STATIC std::vector<Material> Material::materialTable;
STATIC Material Material::s_missingMaterial( "_MissingMaterial_", Rgba( 255, 254, 253 ), -1 );


//-------------------------------------------------------------------------------------------------
// Constructors
//-------------------------------------------------------------------------------------------------
Material::Material( const std::string& name, const Rgba& color, int materialID )
:
m_name( name ),
m_color( color ),
m_mapGfxTileCoords( 0, 0 ),
m_mapGfxTexCoordsTopLeft( Vector2::ZERO ),
m_id( materialID ),
m_flags( 0 ),
m_spawnEntityType( ENT_NONE ),
m_replacementMaterialOnSpawn( 0 )
{
}


Material::Material( const Material& copy )
:
m_name( copy.m_name ),
m_color( copy.m_color ),
m_mapGfxTileCoords( copy.m_mapGfxTileCoords ),
m_mapGfxTexCoordsTopLeft( Vector2::ZERO ),
m_id( copy.m_id ),
m_flags( 0 ),
m_spawnEntityType( ENT_NONE ),
m_replacementMaterialOnSpawn( 0 )
{
}


Material::Material()
:
m_name( "" ),
m_color( Rgba::BLACK ),
m_mapGfxTileCoords( 0, 0 ),
m_mapGfxTexCoordsTopLeft( Vector2::ZERO ),
m_id( 0 ),
m_flags( 0 ),
m_spawnEntityType( ENT_NONE ),
m_replacementMaterialOnSpawn( 0 )
{
}


//-------------------------------------------------------------------------------------------------
// Destructor
//-------------------------------------------------------------------------------------------------
Material::~Material()
{
}



//-------------------------------------------------------------------------------------------------
// ParseColorKeyTable
//-------------------------------------------------------------------------------------------------
STATIC void Material::ParseColorKeyTable( ParseFile& parseFile )
{
	char materialName[ 1024 ];

	char nextChar = parseFile.AdvanceToNextToken();
	while( nextChar != '}' )
	{
		Rgba color;
		color.r = parseFile.ParseInt();
		color.g = parseFile.ParseInt();
		color.b = parseFile.ParseInt();
		color.a = (byte_t) 255;
		parseFile.ParseContiguousString( materialName, sizeof( materialName ) );

		Material newMaterial( materialName, color, g_numMaterials );
//		materialTable.push_back( newMaterial );
		g_materialTable[ g_numMaterials++ ] = newMaterial;
		nextChar = parseFile.AdvanceToNextToken();
	}
}


//-------------------------------------------------------------------------------------------------
// FindMaterialByName
//-------------------------------------------------------------------------------------------------
STATIC Material& Material::FindMaterialByName( const char* name )
{
	for( int i = 0; i < g_numMaterials; i ++ )
	{
		Material& material = g_materialTable[ i ];
		if( !stricmp( material.m_name.c_str(), name ) )
		{
			return material;
		}
	}

	return s_missingMaterial;
	ASSERT_MSG1( false, "Material::FindMaterialByName() : couldn't find material named \"%s\"", name );
}


//-------------------------------------------------------------------------------------------------
// FindMaterialByID
//-------------------------------------------------------------------------------------------------
STATIC Material& Material::FindMaterialByID( int id )
{
	return g_materialTable[ id ];
}


//-------------------------------------------------------------------------------------------------
// FindMaterialByColor
//-------------------------------------------------------------------------------------------------
STATIC Material& Material::FindMaterialByColor( const Rgba& color )
{
	for( int i = 0; i < g_numMaterials; i ++ )
	{
		Material& material = g_materialTable[ i ];
		if( material.m_color == color )
		{
			return material;
		}
	}

	return s_missingMaterial;
	ASSERT_MSG3( false, "Material::FindMaterialByColor() : couldn't find material matching color %d %d %d", color.r, color.g, color.b );
}


//-------------------------------------------------------------------------------------------------
// ParseMaterialData
//-------------------------------------------------------------------------------------------------
STATIC void Material::ParseMaterialData( ParseFile& parseFile, const char* name )
{
	Material& material = FindMaterialByName( name );
	VALIDATE_MSG1( material.m_id != -1, "ParseMaterialData (\"%s\") - no such material found in TileColorKey section!\n", name );

	char keyName[ 1024 ];

	/// Read the key name
	parseFile.ParseContiguousString( keyName, sizeof( keyName ) );
//	theConsole.Printf( "Parsing material key \"%s\" for material \"%s\"\n", keyName, name );

	if( !stricmp( keyName, "mapgfxtilecoords" ) )
	{
		ParseMaterialKey_mapgfxtilecoords( parseFile, material, keyName );
	}
	else if( !stricmp( keyName, "impassable" ) )
	{
		ParseMaterialKey_impassable( parseFile, material, keyName );
	}
	else if( !stricmp( keyName, "blocksplayers" ) )
	{
		ParseMaterialKey_blocksplayers( parseFile, material, keyName );
	}
	else if( !stricmp( keyName, "blocksmonsters" ) )
	{
		ParseMaterialKey_blocksmonsters( parseFile, material, keyName );
	}
	else if( !stricmp( keyName, "blocksprojectiles" ) )
	{
		ParseMaterialKey_blocksprojectiles( parseFile, material, keyName );
	}
	else if( !stricmp( keyName, "spawn" ) )
	{
		ParseMaterialKey_spawn( parseFile, material, keyName );
	}
	else if( !stricmp( keyName, "replacewith" ) )
	{
		ParseMaterialKey_replacewith( parseFile, material, keyName );
	}
	else if( !stricmp( keyName, "event" ) )
	{
		ParseMaterialKey_event( parseFile, material, keyName );
//		material.PrintDebugInfo();
	}
	else
	{
		PANIC_MSG2( "Material::ParseMaterialData() : unknown material key \"%s\" for material \"%s\"", keyName, name );
	}
}


//-------------------------------------------------------------------------------------------------
// ParseMaterialDescription
//-------------------------------------------------------------------------------------------------
STATIC void Material::ParseMaterialDescription( ParseFile& parseFile )
{
	char materialName[ 1024 ];

	parseFile.ParseContiguousString( materialName, sizeof( materialName ) );
	parseFile.ParseExpectedToken( '{' );

	char nextChar = parseFile.AdvanceToNextToken();
	while( nextChar != '}' )
	{
		ParseMaterialData( parseFile, materialName );
		nextChar = parseFile.AdvanceToNextToken();
	}

	parseFile.ParseExpectedToken( '}' );
}


//-------------------------------------------------------------------------------------------------
// BuildMaterialTable
//-------------------------------------------------------------------------------------------------
STATIC void Material::BuildMaterialTable()
{
	/// Build "missing material" material
	s_missingMaterial.m_mapGfxTileCoords.SetXY( 16, 16 );
	int texCoordX = ((s_missingMaterial.m_mapGfxTileCoords.x - 1) * 32);
	int texCoordY = ((s_missingMaterial.m_mapGfxTileCoords.y - 1) * 32);
	s_missingMaterial.m_mapGfxTexCoordsTopLeft.x = (float) texCoordX / 512.0f;
	s_missingMaterial.m_mapGfxTexCoordsTopLeft.y = (float) texCoordY / 512.0f;

	ParseFile parseFile( "data/maps/tiles_general.txt" );
	parseFile.StripComments();

	parseFile.ParseExpectedToken( "TileColorKey" );
	parseFile.ParseExpectedToken( '{' );
	ParseColorKeyTable( parseFile );
	parseFile.ParseExpectedToken( '}' );

	char nextChar = parseFile.AdvanceToNextToken();
	while( nextChar )
	{
		ParseMaterialDescription( parseFile );
		nextChar = parseFile.AdvanceToNextToken();
	}
}


//-------------------------------------------------------------------------------------------------
// Init
//-------------------------------------------------------------------------------------------------
STATIC void Material::Init()
{
	int i;

	// initialize the entity type map; converts string to enum for spawning
	s_entityTypeMap.clear();
	s_entityTypeMap[ "PlayerStart" ]			= ENT_PLAYER_START;
	s_entityTypeMap[ "GeneratorGrunt" ]			= ENT_GENERATOR_GRUNT;
	s_entityTypeMap[ "GeneratorMushroom" ]		= ENT_GENERATOR_MUSHROOM;
	s_entityTypeMap[ "GeneratorWizrobe" ]		= ENT_GENERATOR_WIZROBE;
	s_entityTypeMap[ "GeneratorBoss1" ]			= ENT_GENERATOR_BOSS1;
	s_entityTypeMap[ "GeneratorSingleBoss1" ]	= ENT_GENERATOR_SINGLEBOSS1;
	s_entityTypeMap[ "GeneratorBoss2" ]			= ENT_GENERATOR_BOSS2;
	s_entityTypeMap[ "GeneratorSingleBoss2" ]	= ENT_GENERATOR_SINGLEBOSS2;
	s_entityTypeMap[ "EnemyGrunt" ]				= ENT_ENEMY_GRUNT;
	s_entityTypeMap[ "EnemyMushroom" ]			= ENT_ENEMY_MUSHROOM;
	s_entityTypeMap[ "EnemyWizrobe" ]			= ENT_ENEMY_WIZROBE;
	s_entityTypeMap[ "EnemyBoss1" ]				= ENT_ENEMY_BOSS1;
	s_entityTypeMap[ "EnemyBoss2" ]				= ENT_ENEMY_BOSS2;
	s_entityTypeMap[ "Barrel" ]					= ENT_BARREL;
	s_entityTypeMap[ "Crate" ]					= ENT_CRATE;

	BuildMaterialTable();

	/// Report all materials
	theConsole.Printf( "Total of %d materials types known:\n", g_numMaterials );
	for( i = 0; i < g_numMaterials; i ++ )
	{
		Material& mat = g_materialTable[ i ];

//		theConsole.Printf( "Material #%3d : (%03d %03d %03d) (%2d,%2d) : \"%s\"\n",
//			i,
//			mat.m_color.r, mat.m_color.g, mat.m_color.b,
//			mat.m_mapGfxTileCoords.x, mat.m_mapGfxTileCoords.y,
//			mat.m_name.c_str() );

	}
}


//-------------------------------------------------------------------------------------------------
// ParseMaterialKey_mapgfxtilecoords
//-------------------------------------------------------------------------------------------------
STATIC void Material::ParseMaterialKey_mapgfxtilecoords( ParseFile& parseFile, Material& material, const char* keyName )
{
	material.m_mapGfxTileCoords.x = parseFile.ParseInt();
	parseFile.ParseExpectedToken( ',' );
	material.m_mapGfxTileCoords.y = parseFile.ParseInt();

	/// Generate float tex coords
	int texCoordX = ((material.m_mapGfxTileCoords.x - 1) * 32);
	int texCoordY = ((material.m_mapGfxTileCoords.y - 1) * 32);
	material.m_mapGfxTexCoordsTopLeft.x = (float) texCoordX / 512.0f;
	material.m_mapGfxTexCoordsTopLeft.y = (float) texCoordY / 512.0f;
}


//-------------------------------------------------------------------------------------------------
// ParseMaterialKey_impassable
//-------------------------------------------------------------------------------------------------
STATIC void Material::ParseMaterialKey_impassable( ParseFile& parseFile, Material& material, const char* keyName )
{
//	theConsole.Printf( "Read IMPASSABLE (BLOCKSPLAYERS + BLOCKSMONSTERS) flag for material %s\n", material.m_name.c_str() );
	material.m_flags |= (MATFLAG_BLOCKS_PLAYERS | MATFLAG_BLOCKS_ENEMIES | MATFLAG_BLOCKS_PROJECTILES);
}


//-------------------------------------------------------------------------------------------------
// ParseMaterialKey_blocksplayers
//-------------------------------------------------------------------------------------------------
STATIC void Material::ParseMaterialKey_blocksplayers( ParseFile& parseFile, Material& material, const char* keyName )
{
//	theConsole.Printf( "Read BLOCKSPLAYERS flag for material %s\n", material.m_name.c_str() );
	material.m_flags |= (MATFLAG_BLOCKS_PLAYERS);
}


//-------------------------------------------------------------------------------------------------
// ParseMaterialKey_blocksmonsters
//-------------------------------------------------------------------------------------------------
STATIC void Material::ParseMaterialKey_blocksmonsters( ParseFile& parseFile, Material& material, const char* keyName )
{
//	theConsole.Printf( "Read BLOCKSMONSTERS flag for material %s\n", material.m_name.c_str() );
	material.m_flags |= (MATFLAG_BLOCKS_ENEMIES);
}


//-------------------------------------------------------------------------------------------------
// ParseMaterialKey_blocksprojectiles
//-------------------------------------------------------------------------------------------------
STATIC void Material::ParseMaterialKey_blocksprojectiles( ParseFile& parseFile, Material& material, const char* keyName )
{
//	theConsole.Printf( "Read BLOCKSMONSTERS flag for material %s\n", material.m_name.c_str() );
	material.m_flags |= (MATFLAG_BLOCKS_PROJECTILES);
}


//-------------------------------------------------------------------------------------------------
// ParseMaterialKey_spawn
//-------------------------------------------------------------------------------------------------
STATIC void Material::ParseMaterialKey_spawn( ParseFile& parseFile, Material& material, const char* keyName )
{
	theConsole.Printf( "Read SPAWN command for material %s\n", material.m_name.c_str() );

	char entityType[ 128 ];
	parseFile.ParseContiguousString( entityType, sizeof( entityType ) );
	if( s_entityTypeMap.end() != s_entityTypeMap.find( entityType ) )
	{
		material.m_spawnEntityType = s_entityTypeMap[ entityType ];
	}
}

//-------------------------------------------------------------------------------------------------
// ParseMaterialKey_replacewith
//-------------------------------------------------------------------------------------------------
STATIC void Material::ParseMaterialKey_replacewith( ParseFile& parseFile, Material& material, const char* keyName )
{
//	theConsole.Printf( "Read REPLACEWITH flag for material %s\n", material.m_name.c_str() );
	
	char materialName[ 128 ];
	parseFile.ParseContiguousString( materialName, sizeof( materialName ) );
	material.m_replacementMaterialOnSpawn = strdup( materialName );
}


//-------------------------------------------------------------------------------------------------
// ParseMaterialKey_event
//-------------------------------------------------------------------------------------------------
STATIC void Material::ParseMaterialKey_event( ParseFile& parseFile, Material& material, const char* keyName )
{
//	theConsole.Printf( "Read EVENT command for material %s\n", material.m_name.c_str() );

	/// Parse the event type name
	char eventTypeName[ 128 ];
	parseFile.ParseContiguousString( eventTypeName, sizeof( eventTypeName ) );

	/// Build the new event handler locally before adding it to the material
	ScriptEventHandler newEventHandler;
	newEventHandler.m_eventTypeName = eventTypeName;
	
	/// Parse the action block for this event
	parseFile.ParseExpectedToken( '{' );
	char nextChar = parseFile.AdvanceToNextToken();
	while( nextChar != '}' )
	{
		ParseEventAction( parseFile, newEventHandler );
		nextChar = parseFile.AdvanceToNextToken();
	}
	parseFile.ParseExpectedToken( '}' );

	/// Add the new event handler to this material
	material.m_eventsHandled.push_back( newEventHandler );
}


//-------------------------------------------------------------------------------------------------
// ParseEventAction
//-------------------------------------------------------------------------------------------------
STATIC void Material::ParseEventAction( ParseFile& parseFile, ScriptEventHandler& eventHandler )
{
	/// Parse the action name
	char actionName[ 128 ];
	parseFile.ParseContiguousString( actionName, sizeof( actionName ) );

	/// Parse the remainder of the line for the action
	char actionArgs[ 1024 ];
	strcpy( actionArgs, "" );
	char nextToken = parseFile.AdvanceToNextToken();
	if( nextToken != '}' )
	{
		parseFile.ParseStringToEndOfLine( actionArgs, sizeof( actionArgs ) );
	}

	/// Build a new event action item
	ScriptAction newAction;
	newAction.m_actionName = actionName;
	newAction.m_actionArgs = actionArgs;

	/// Add it to the specified event handler's action list
	eventHandler.m_scriptActions.push_back( newAction );
}



//-------------------------------------------------------------------------------------------------
// PrintDebugInfo
//-------------------------------------------------------------------------------------------------
void Material::PrintDebugInfo()
{
	theConsole.Printf( "Material Debug Info for \"%s\"\n", m_name.c_str() );
	theConsole.Printf( "Color is %d %d %d\n", m_color.r, m_color.g, m_color.b );
	theConsole.Printf( "MapGfxTileCoords are %d,%d\n", m_mapGfxTileCoords.x, m_mapGfxTileCoords.y );
	theConsole.Printf( "Material ID # is %d\n", m_id );
	theConsole.Printf( "Flags == %d (0x%08x)\n", m_flags, m_flags );

	theConsole.Printf( "Total of %d event handlers:\n", m_eventsHandled.size() );
	std::list<ScriptEventHandler>::iterator iter;
	for( iter = m_eventsHandled.begin(); iter != m_eventsHandled.end(); ++ iter )
	{
		ScriptEventHandler& event = *iter;
		theConsole.Printf( "Event handler for \"%s\":\n", event.m_eventTypeName.c_str() );

		std::list<ScriptAction>::iterator aIter;
		for( aIter = event.m_scriptActions.begin(); aIter != event.m_scriptActions.end(); ++ aIter )
		{
			ScriptAction& action = *aIter;
			theConsole.Printf( "  %s %s\n", action.m_actionName.c_str(), action.m_actionArgs.c_str() );
		}
	}

	theConsole.Printf( "\n" );
/*
	std::string		m_name;
	Rgba			m_color;
	IntVector2		m_mapGfxTileCoords;
	Vector2			m_mapGfxTexCoordsTopLeft;
	int				m_id;
	unsigned int	m_flags;
	TypeOfEntity	m_spawnEntityType;
	char*			m_replacementMaterialOnSpawn;
	std::list<ScriptEventHandler>	m_eventsHandled;
*/
	
}

//-------------------------------------------------------------------------------------------------
// GetEventHandlerForEvent
//-------------------------------------------------------------------------------------------------
ScriptEventHandler* Material::GetEventHandlerForEvent( std::string eventTypeName )
{
	std::list<ScriptEventHandler>::iterator iter;
	for( iter = m_eventsHandled.begin(); iter != m_eventsHandled.end(); ++ iter )
	{
		ScriptEventHandler& event = *iter;
		if( event.m_eventTypeName == eventTypeName )
		{
			return &event;
		}
	}

	return NULL;
}



