//-------------------------------------------------------------------------------------------------
// Scripting.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "Scripting.hpp"
#include "engineAssert.hpp"
#include "TheGame.hpp"
#include "Material.hpp"

using namespace std;


//-------------------------------------------------------------------------------------------------
// ExecuteAction_Teleport
//-------------------------------------------------------------------------------------------------
void ExecuteAction_Teleport( std::string args )
{
	ParseBuffer parseArgs( args );
	char targetName[ 1024 ];
	parseArgs.ParseContiguousString( targetName, 1024 );
	int levelChangeDelta = parseArgs.ParseInt();

	theGame->m_player->Teleport( targetName, levelChangeDelta );
}


//-------------------------------------------------------------------------------------------------
// ExecuteAction_Trigger
//-------------------------------------------------------------------------------------------------
void ExecuteAction_Trigger( const IntVector2& tilePosAtPointOfExecution, std::string args )
{
	theConsole.Printf( "Triggering materials of type \"%s\" from point %d,%d\n", args.c_str(), tilePosAtPointOfExecution.x, tilePosAtPointOfExecution.y );

	/// Find material id matching <args>
	Material& material = Material::FindMaterialByName( args.c_str() );
	if( material.m_id == -1 )
		return;

	/// Scan all tiles in the current map for a matching material id
	int id = material.m_id;
	MapData* mapData = theGame->m_currentMap;
	for( int x = 0; x < MAP_TILE_SIZE_X; x ++ )
	for( int y = 0; y < MAP_TILE_SIZE_Y; y ++ )
	{
		if( mapData->tileArray[ x ][ y ].m_materialID == id )
		{
			Material& scanMat = Material::FindMaterialByID( id );

			/// For each match, invoke "OnTrigger"
			ScriptEventHandler* onTrigger = scanMat.GetEventHandlerForEvent( "OnTrigger" );
			if( onTrigger )
			{
				ExecuteScriptActions( IntVector2( x, y ), onTrigger->m_scriptActions );
			}
		}
	}
}


///------------------------------------------------------------------------------------------------
/// ExecuteAction_ChangeMaterial
///------------------------------------------------------------------------------------------------
void ExecuteAction_ChangeMaterial( const IntVector2& tilePosAtPointOfExecution, std::string& args )
{
	theConsole.Printf( "Changing material of location %d,%d to type \"%s\"\n", tilePosAtPointOfExecution.x, tilePosAtPointOfExecution.y, args.c_str() );
	MapData* mapData = theGame->m_currentMap;
	if ( mapData )
	{
		mapData->ChangeMaterialAtPosition( tilePosAtPointOfExecution, args );
	}
}


///------------------------------------------------------------------------------------------------
/// ExecuteAction_Give
///------------------------------------------------------------------------------------------------
void ExecuteAction_Give( const IntVector2& tilePosAtPointOfExecution, std::string& args )
{
	theConsole.Printf( "Giving item(s): \"%s\"\n", args.c_str() );

	char itemName[ 256 ];
	int itemCount = 1;
	int argsMet = sscanf( args.c_str(), "%s %d", &itemName, &itemCount );
	if( argsMet != 2 )
	{
		strcpy( itemName, args.c_str() );
	}

	theGame->m_player->m_inventory.GiveItem( itemName, itemCount );
}


///------------------------------------------------------------------------------------------------
/// ExecuteAction_TryUnlock
///------------------------------------------------------------------------------------------------
void ExecuteAction_TryUnlock( const IntVector2& tilePosAtPointOfExecution, std::string& args )
{
	theConsole.Printf( "Trying to unlock at %d,%d\n", tilePosAtPointOfExecution.x, tilePosAtPointOfExecution.y );
	MapData* mapData = theGame->m_currentMap;
	if( !mapData )
	{
		return;
	}

	/// Check if the player has any keys
	if( theGame->m_player->m_inventory.m_numKey <= 0 )
	{
		return;
	}

	/// Use a key
	theGame->m_player->m_inventory.m_numKey --;

	/// Trigger the door
	int x = tilePosAtPointOfExecution.x;
	int y = tilePosAtPointOfExecution.y;
	int materialID = mapData->tileArray[ x ][ y ].m_materialID;
	Material& material = Material::FindMaterialByID( materialID );
	ScriptEventHandler* onTrigger = material.GetEventHandlerForEvent( "OnTrigger" );
	if( onTrigger )
	{
		ExecuteScriptActions( IntVector2( x, y ), onTrigger->m_scriptActions );
	}
}


///-----------------------------------------------------------------------------------------------
/// ExecutionAction_DamagePlayer
///-----------------------------------------------------------------------------------------------
void ExecuteAction_DamagePlayer( const IntVector2& tilePosAtPointOfExecution, std::string& args )
{
	float healthDelta = (float) atof( args.c_str() );
	theGame->m_player->AdjustHealth( healthDelta * -1.0f );
}


//-------------------------------------------------------------------------------------------------
// ExecuteScriptAction
//-------------------------------------------------------------------------------------------------
void ExecuteScriptAction( const IntVector2& tilePosAtPointOfExecution, ScriptAction& action )
{
	theConsole.Printf( "Executing script action:  %s %s  from location %d,%d\n", action.m_actionName.c_str(), action.m_actionArgs.c_str(), tilePosAtPointOfExecution.x, tilePosAtPointOfExecution.y );

	if( action.m_actionName == "Teleport" )
	{
		ExecuteAction_Teleport( action.m_actionArgs );
	}
	else if ( action.m_actionName == "Trigger" )
	{
		ExecuteAction_Trigger( tilePosAtPointOfExecution, action.m_actionArgs );
	}
	else if ( action.m_actionName == "ChangeMaterial" )
	{
		ExecuteAction_ChangeMaterial( tilePosAtPointOfExecution, action.m_actionArgs );
	}
	else if ( action.m_actionName == "Damage" )
	{
		ExecuteAction_DamagePlayer( tilePosAtPointOfExecution, action.m_actionArgs );
	}
	else if ( action.m_actionName == "TryUnlock" )
	{
		ExecuteAction_TryUnlock( tilePosAtPointOfExecution, action.m_actionArgs );
	}
	else if ( action.m_actionName == "Give" )
	{
		ExecuteAction_Give( tilePosAtPointOfExecution, action.m_actionArgs );
	}
}


//-------------------------------------------------------------------------------------------------
// ExecuteScriptActions
//-------------------------------------------------------------------------------------------------
void ExecuteScriptActions( const IntVector2& tilePosAtPointOfExecution, std::list<ScriptAction>& actionList )
{
	std::list<ScriptAction>::iterator iter;
	for( iter = actionList.begin(); iter != actionList.end(); ++ iter )
	{
		ScriptAction& action = *iter;
		ExecuteScriptAction( tilePosAtPointOfExecution, action );
	}
}


//-------------------------------------------------------------------------------------------------
// DamageMapGridXY
//-------------------------------------------------------------------------------------------------
void DamageMapGridXY( const IntVector2& location , bool isBomb )
{
	unsigned int frameNum = theApplication.GetFrameNumber();
	MapData* mapData = MapData::s_mapDataList[ theGame->m_currentMapNumber ] ;

	int x = location.x;
	int y = location.y;

	mapData->tileArray[ x ][ y ].m_debugFrameCountMark = frameNum + 10;

	int materialID = mapData->tileArray[ x ][ y ].m_materialID;
	Material& material = Material::FindMaterialByID( materialID );
	if( isBomb == true )
	{
		ScriptEventHandler* onBombDamage = material.GetEventHandlerForEvent( "OnBombDamage" );
		if( onBombDamage )
		{
			ExecuteScriptActions( location, onBombDamage->m_scriptActions );
		}
		ScriptEventHandler* onDamage = material.GetEventHandlerForEvent( "OnDamage" );
		if( onDamage )
		{
			ExecuteScriptActions( location, onDamage->m_scriptActions );
		}
	}
	else
	{
		ScriptEventHandler* onDamage = material.GetEventHandlerForEvent( "OnDamage" );
		if( onDamage )
		{
			ExecuteScriptActions( location, onDamage->m_scriptActions );
		}
	}
}

//-------------------------------------------------------------------------------------------------
// DamageMapGridRadius
//-------------------------------------------------------------------------------------------------
void DamageMapGridRadius( const Vector2& center, float radius, bool isBomb )
{
	theConsole.Printf( "Causing RADIAL damage at: %.2f,%.2f, radius == %.2f\n", center.x, center.y, radius );

	Vector2 radiusVector( radius, radius );
	Vector2 mins = center - radiusVector;
	Vector2 maxs = center + radiusVector;
	IntVector2 intMins( (int)mins.x, (int)mins.y );
	IntVector2 intMaxs( (int)maxs.x, (int)maxs.y );
	IntVector2 mapMins( 0, 0 );
	IntVector2 mapMaxs( MAP_TILE_SIZE_X, MAP_TILE_SIZE_Y );
	intMins.Clamp( mapMins, mapMaxs );

	float radiusSquared = radius * radius;
	for( int x = intMins.x; x <= intMaxs.x; x ++ )
	for( int y = intMins.y; y <= intMaxs.y; y ++ )
	{
		Vector2 squareCenter( 0.5f + (float)x, 0.5f + (float)y );
		float distSquared = DistanceSquaredBetweenPoints2d( squareCenter, center );
		if( distSquared <= radiusSquared )
		{
			if( isBomb == true )
			{	
				DamageMapGridXY( IntVector2( x, y ) , true );
			}
			else
			{
				DamageMapGridXY( IntVector2( x, y ) , false );
			}
		}
	}
}



//-------------------------------------------------------------------------------------------------
// DamageEntitiesRadius
//-------------------------------------------------------------------------------------------------
void DamageEntitiesRadius( const Vector2& center, float radius, float maxDamage )
{
	MapData* mapdata = theGame->m_currentMap;

	std::list< Entity* >::iterator start = mapdata->entityList.begin();
	std::list< Entity* >::iterator end = mapdata->entityList.end();
	std::list< Entity* >::iterator iter;
	for( iter = start; iter != end; ++ iter )
	{
		Entity* entity = (*iter);

		float entitySizeX = entity->m_bounds.maxs.x - entity->m_bounds.mins.x;
		float entitySizeY = entity->m_bounds.maxs.y - entity->m_bounds.mins.y;
		float entityDiameter = (entitySizeX + entitySizeY) * 0.5f * (1.0f / 32.0f);
		float entityRadius = (entitySizeX + entitySizeY) * 0.5f * (1.0f / 32.0f);

		float effectiveRadius = radius + entityRadius;
		float radiusSquared = effectiveRadius * effectiveRadius;
		float distSquared = DistanceSquaredBetweenPoints2d( center, entity->m_pos );

		if( distSquared < radiusSquared )
		{
			theConsole.Printf( "radius = %.2f, effectiveRadius = %.2f, distSquared = %.2f   radiusSquared = %.2f\n", radius, effectiveRadius, distSquared, radiusSquared );

			float realDistance = (float) sqrt( distSquared );
			float realRadius = (float) sqrt( radiusSquared );
			float distanceFraction = realDistance / realRadius;
			distanceFraction = ClampFloat( distanceFraction, 0.0f, 1.0f );
			float damageFraction = 1.0f - distanceFraction;
			float damageAmount = damageFraction * maxDamage;

			theConsole.Printf( "Dealing %.0f damage to entity type #%d at %.0f,%.0f (distance = %.2f)\n",
				damageAmount, entity->m_type, entity->m_pos.x, entity->m_pos.y, realDistance );

			//HACK -- Hack for Boss monsters
			if( entity->m_type != ENT_ENEMY_BOSS1 )
			{
				if( damageAmount == 0 && entity->m_isEnemy == true )
				{
					entity->m_stunned = true;
				}
				else
				{
					if( entity->m_type == ENT_ENEMY_BOSS2 )
					{
						entity->DealDamage( damageAmount * 0.5 );
						entity->m_isBeingDamagedByPlayer = true;
					}
					else
					{
						entity->DealDamage( damageAmount );
						entity->m_isBeingDamagedByPlayer = true;
					}
				}
			}
		}
	}
}


