//-------------------------------------------------------------------------------------------------
// TheGame.cpp (game)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
// 
// This source may not be used, modified, or distributed
//  without the express written permission of Brian 'Squirrel' Eiserloh.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "TheGame.hpp"
#include "TheEditor.hpp"


//-------------------------------------------------------------------------------------------------
// Globals
//-------------------------------------------------------------------------------------------------
TheGame* theGame = NULL;


//-------------------------------------------------------------------------------------------------
// Destructor()
//-------------------------------------------------------------------------------------------------
TheGame::~TheGame()
{
}


//-------------------------------------------------------------------------------------------------
// Constructor
//-------------------------------------------------------------------------------------------------
TheGame::TheGame()
:
m_currentMapNumber( 1 ),
m_noClip( false ),
m_player( NULL ),
m_gameState( GAMESTATE_ATTRACT ),
m_attractScreenTexture( NULL ),
m_gameOverScreenTexture( NULL ),
m_youWinScreenTexture( NULL ),
_commandConsoleEnabled( false ),
_commandBufferPosition(0)
{
	strcpy( _commandBuffer, "" );
}


//-------------------------------------------------------------------------------------------------
// PreInit
//-------------------------------------------------------------------------------------------------
void TheGame::PreInit()
{
	/// Assign application name
	theApplication.SetName( "NetSlack v0.02" );

	/// Choose a video mode
	SHORT result = GetAsyncKeyState( VK_NUMLOCK );
	if( result & 0x8000 )
		theVideoSystem.SetDefaultVidModeProperties( 1024, 768, 32, 85, TheVideoSystem::VIDMODE_FLAG_FULLSCREEN );
	else
		theVideoSystem.SetDefaultVidModeProperties( 1024, 768, 32, 85, 0 );
}

#include "IntVector4.hpp"

//-------------------------------------------------------------------------------------------------
// PostInit
//-------------------------------------------------------------------------------------------------
void TheGame::PostInit()
{
	Material::Init();

	m_lights.setTexture( new Texture( "data/lights/attenuation.tga" ) );

	// Create the player
	m_player = new Player();
	VALIDATE( m_player );

	_InitMap();

	m_attractScreenTexture = new Texture( "data/gfx/title.tga" );
	m_gameOverScreenTexture = new Texture( "data/gfx/gameover.tga" );
	m_youWinScreenTexture = new Texture( "data/gfx/win.tga" );

	WarpToLevel( 1 );
}


//-------------------------------------------------------------------------------------------------
void TheGame::_InitMap()
{
	_mapDisplay.LoadTiles( "mapgfx001" );
	WarpToLevelSet( "charon" );
}


//-------------------------------------------------------------------------------------------------
// RunFrame_Playing
//-------------------------------------------------------------------------------------------------
void TheGame::RunFrame_Playing()
{
	/// Process input, update entities, run game logic
	UpdateEverything();

	/// Render the scene
	DrawEverything();
}


//-------------------------------------------------------------------------------------------------
// RunFrame_Attract
//-------------------------------------------------------------------------------------------------
void TheGame::RunFrame_Attract()
{
	SetUpScreenCoordinateSystem();
	theRenderer.DrawTexturedRect( *m_attractScreenTexture, Vector2::ZERO, Vector2( 1024, 768 ), Vector2::ZERO, Vector2( 1.0f, 0.75f ) );
}


//-------------------------------------------------------------------------------------------------
// RunFrame_GameOver
//-------------------------------------------------------------------------------------------------
void TheGame::RunFrame_GameOver()
{
	SetUpScreenCoordinateSystem();
	theRenderer.DrawTexturedRect( *m_gameOverScreenTexture, Vector2::ZERO, Vector2( 1024, 768 ), Vector2::ZERO, Vector2( 1.0f, 0.75f ) );
}


//-------------------------------------------------------------------------------------------------
// RunFrame_YouWin
//-------------------------------------------------------------------------------------------------
void TheGame::RunFrame_YouWin()
{
	SetUpScreenCoordinateSystem();
	theRenderer.DrawTexturedRect( *m_youWinScreenTexture, Vector2::ZERO, Vector2( 1024, 768 ), Vector2::ZERO, Vector2( 1.0f, 0.75f ) );
}


//-------------------------------------------------------------------------------------------------
// RunFrame_EditMap
//-------------------------------------------------------------------------------------------------
void TheGame::RunFrame_EditMap()
{
	if( !theEditor )
	{
		m_gameState = GAMESTATE_ATTRACT;
		return;
	}

	theEditor->RunFrame();
}


//-------------------------------------------------------------------------------------------------
// RunFrame
//-------------------------------------------------------------------------------------------------
void TheGame::RunFrame()
{
	switch( m_gameState )
	{
		case GAMESTATE_ATTRACT:		RunFrame_Attract();		break;
		case GAMESTATE_PLAYING:		RunFrame_Playing();		break;
		case GAMESTATE_GAMEOVER:	RunFrame_GameOver();	break;
		case GAMESTATE_YOUWIN:		RunFrame_YouWin();		break;
		case GAMESTATE_EDITMAP:		RunFrame_EditMap();		break;

		default:
			ChangeState( GAMESTATE_ATTRACT );
			break;
	}
}


//-------------------------------------------------------------------------------------------------
// WarpToLevel
//-------------------------------------------------------------------------------------------------
bool TheGame::WarpToLevel( unsigned int newMapNumber, int newTileX, int newTileY )
{
	int size = MapData::s_mapDataList.size();

	/// Don't go out of bounds of loaded map numbers
	if( (newMapNumber < 1) || (newMapNumber > MapData::s_mapDataList.size() ) )
		return false;

	m_currentMapNumber = newMapNumber - 1;
	m_currentMap = MapData::s_mapDataList[ m_currentMapNumber ];

	m_lights.m_mapRadiusScale = m_currentMap->m_lightScale;
	m_lights.m_fogColor = m_currentMap->m_fogColor;

	if( newTileX != -1 && newTileY != -1 )
	{
		m_player->pos.x = newTileX + 0.5f;
		m_player->pos.y = newTileY + 0.5f;
	}
	else
	{
		std::list< Entity* >::iterator start = m_currentMap->entityList.begin();
		std::list< Entity* >::iterator end = m_currentMap->entityList.end();
		
		for( std::list< Entity* >::iterator cur = start; cur != end; cur++ )
		{
			if( (*cur)->m_type == ENT_PLAYER_START )
			{
				m_player->pos.x = 0.5f + (*cur)->m_pos.x;
				m_player->pos.y = 0.5f + (*cur)->m_pos.y;
			}
		}
	}

	return true;
}


///-------------------------------------------------------------------------------------------------
/// WarpToLevelSet -- This clears out the current level set, and then loads
/// the new set.  It then warps to the specified level.
///--------------------------------------------------------------------------------------------------
bool TheGame::WarpToLevelSet( const char* mapSet, unsigned int newMapNumber, int newTileX, int newTileY )
{
	MapData::s_mapDataList.erase( MapData::s_mapDataList.begin(), MapData::s_mapDataList.end() );

	/// Load as many maps as exist
	bool finished = false;
	for( int mapNumber = 1; !finished ; mapNumber++ )
	{
		finished = !MapData::LoadMap( mapSet, mapNumber );
	}

	WarpToLevel( newMapNumber, newTileX, newTileY );
	return true ;
}


//-------------------------------------------------------------------------------------------------
// ChangeState
//-------------------------------------------------------------------------------------------------
void TheGame::ChangeState( GameState newState )
{
	if( m_gameState == newState )
		return;

	if( newState == GAMESTATE_PLAYING )
	{
		m_player->health = 100.0f;
	}

	m_gameState = newState;
}




